#include <QSocketNotifier>
#include <QDeclarativeItem>
#include <QGraphicsScene>
#include <QGraphicsView>
#include <QGraphicsWidget>
#include <typeinfo>

#include "geissingleton.h"
#include "gesturearea.h"
#include "gesture.h"

/******************************************************************************
 * Template functions for getting object attributes from GEIS
 ******************************************************************************/

template<typename Object>
static GeisAttr getAttributeValue(Object object, const char *attributeName)
{
    qCritical("Unhandled attribute value getter");
    throw;
}

#define DEFINE_ATTRIBUTE_VALUE(OBJECT, FUNC) \
    template<> \
    GeisAttr getAttributeValue(OBJECT object, const char *attributeName) \
    { \
        GeisAttr attribute = geis_##FUNC##_attr_by_name(object, attributeName); \
        if (!attribute) { \
            qCritical("Failed to get attribute value for '%s' from " #FUNC, attributeName); \
            throw; \
        } \
        return attribute; \
    }

DEFINE_ATTRIBUTE_VALUE(GeisEvent, event)
DEFINE_ATTRIBUTE_VALUE(GeisFrame, frame)
DEFINE_ATTRIBUTE_VALUE(GeisTouch, touch)

#undef DEFINE_ATTRIBUTE_VALUE

template<typename Ret>
static Ret getAttributeReturn(GeisAttr attribute)
{
    Ret object = static_cast<Ret>(geis_attr_value_to_pointer(attribute));
    if (!object) {
        qCritical("Feailed to get valid %s pointer from attribute '%s'", typeid(Ret).name(), geis_attr_name(attribute));
        throw;
    }

    return object;
}

#define DEFINE_BASIC_ATTRIBUTE_RETURN(RET, FUNC) \
    template<> \
    RET getAttributeReturn(GeisAttr attribute) \
    { \
        return geis_attr_value_to_##FUNC(attribute); \
    }

DEFINE_BASIC_ATTRIBUTE_RETURN(GeisBoolean, boolean)
DEFINE_BASIC_ATTRIBUTE_RETURN(GeisInteger, integer)
DEFINE_BASIC_ATTRIBUTE_RETURN(GeisFloat, float)
DEFINE_BASIC_ATTRIBUTE_RETURN(GeisString, string)

#undef DEFINE_BASIC_ATTRIBUTE_RETURN

template<typename Ret, typename Object>
Ret GeisSingleton::getAttribute(Object object, const char *attributeName)
{
    GeisAttr attribute = getAttributeValue<Object>(object, attributeName);
    return getAttributeReturn<Ret>(attribute);
}

/******************************************************************************
 * Functions for handling the GEIS singleton instance
 ******************************************************************************/

Q_GLOBAL_STATIC(GeisSingleton, _geis_singleton)

GeisSingleton *GeisSingleton::instance() {
    return _geis_singleton();
}

GeisSingleton::GeisSingleton(QObject *parent) :
    QObject(parent),
    _initialized(false)
{
    int fd;

    _geis = geis_new(GEIS_INIT_UTOUCH_XCB, GEIS_INIT_TRACK_DEVICES, GEIS_INIT_TRACK_GESTURE_CLASSES, NULL);
    if (!_geis) {
        qCritical("Failed to initialize geis instance");
    }

    geis_get_configuration(_geis, GEIS_CONFIGURATION_FD, &fd);

    _notifier = new QSocketNotifier(fd, QSocketNotifier::Read, this);
    connect(_notifier, SIGNAL(activated(int)), SLOT(geisEvent()));
}

/******************************************************************************/

void GeisSingleton::registerGestureArea(const GestureArea *area)
{
    if (!_windowMap.contains(area->getWindowId(), area))
        _windowMap.insert(area->getWindowId(), area);
}

void GeisSingleton::geisEvent()
{
    GeisEvent event;
    GeisStatus status;

    status = geis_dispatch_events(_geis);
    if (status != GEIS_STATUS_SUCCESS) {
        qWarning("Failed to dispatch geis events");
        return;
    }

    for (status = geis_next_event(_geis, &event);
         status == GEIS_STATUS_SUCCESS || status == GEIS_STATUS_CONTINUE;
         status = geis_next_event(_geis, &event)) {
        switch (geis_event_type(event)) {
        case GEIS_EVENT_CLASS_AVAILABLE:
        case GEIS_EVENT_CLASS_CHANGED:
        case GEIS_EVENT_CLASS_UNAVAILABLE:
            handleClassEvent(event);
            break;

        case GEIS_EVENT_DEVICE_AVAILABLE:
        case GEIS_EVENT_DEVICE_UNAVAILABLE:
            handleDeviceEvent(event);
            break;

        case GEIS_EVENT_INIT_COMPLETE:
            _initialized = true;
            emit initialized();
            break;

        case GEIS_EVENT_GESTURE_BEGIN:
            handleGestureBeginEvent(event);
            break;
        case GEIS_EVENT_GESTURE_UPDATE:
            handleGestureUpdateEvent(event, false);
            break;
        case GEIS_EVENT_GESTURE_END:
            handleGestureUpdateEvent(event, true);
            break;

        default:
            break;
        }

        geis_event_delete(event);
    }
}

void GeisSingleton::handleClassEvent(GeisEvent event)
{
    GeisGestureClass gestureClass;
    GestureArea::Primitive primitive;

    gestureClass = getAttribute<GeisGestureClass, GeisEvent>(event, GEIS_EVENT_ATTRIBUTE_CLASS);

    GeisString name = geis_gesture_class_name(gestureClass);
    if (!name) {
        qCritical("Failed to get name of gesture class");
        return;
    }

    if (QString::compare(name, GEIS_GESTURE_DRAG) == 0)
        primitive = GestureArea::Drag;
    else if (QString::compare(name, GEIS_GESTURE_PINCH) == 0)
        primitive = GestureArea::Pinch;
    else if (QString::compare(name, GEIS_GESTURE_ROTATE) == 0)
        primitive = GestureArea::Rotate;
    else if (QString::compare(name, GEIS_GESTURE_TAP) == 0)
        primitive = GestureArea::Tap;
    else
        return;

    switch (geis_event_type(event)) {
    case GEIS_EVENT_CLASS_AVAILABLE:
        geis_gesture_class_ref(gestureClass);
        _classes[primitive] = gestureClass;
        break;
    case GEIS_EVENT_CLASS_UNAVAILABLE:
        _classes.remove(primitive);
        geis_gesture_class_unref(gestureClass);
        break;
    default:
        break;
    }
}

void GeisSingleton::handleDeviceEvent(GeisEvent event) {
    GeisDevice device;
    bool direct = false;
    bool independent = false;
    Gesture::DeviceType type;

    device = getAttribute<GeisDevice, GeisEvent>(event, GEIS_EVENT_ATTRIBUTE_DEVICE);

    if (geis_event_type(event) == GEIS_EVENT_DEVICE_AVAILABLE) {
        for (GeisSize i = 0; i < geis_device_attr_count(device); i++) {
            GeisAttr attr = geis_device_attr(device, i);

            if (QString::compare(geis_attr_name(attr), GEIS_DEVICE_ATTRIBUTE_DIRECT_TOUCH) == 0)
                direct = geis_attr_value_to_boolean(attr);
            else if (QString::compare(geis_attr_name(attr), GEIS_DEVICE_ATTRIBUTE_INDEPENDENT_TOUCH) == 0)
                independent = geis_attr_value_to_boolean(attr);
        }

        if (direct && !independent)
            type = Gesture::TouchScreen;
        else if (!direct && !independent)
            type = Gesture::TouchPad;
        else if (!direct)
            type = Gesture::Independent;
        else {
            qWarning("Unknown device type (independent and direct)");
            return;
        }

        _deviceTypeMap[geis_device_id(device)] = type;
    } else /* GEIS_EVENT_DEVICE_UNAVAILABLE */
        _deviceTypeMap.remove(geis_device_id(device));
}

/* Get a list of gesture frames from the groupset. The groupset mechanism
 * doesn't really work, so we flatten everything down into a list. The list
 * must be in order from most touches to least.
 */
void GeisSingleton::getGestureFrameList(GeisGroupSet groupSet, QList<GeisFrame> &frames)
{
    QMultiMap<unsigned int, GeisFrame> framesByTouchCount;
    QList<GeisFrame> handledGestures;

    for (GeisSize i = 0; i < geis_groupset_group_count(groupSet); i++) {
        GeisGroup group = geis_groupset_group(groupSet, i);

        for (GeisSize j = 0; j < geis_group_frame_count(group); j++) {
            GeisFrame frame = geis_group_frame(group, j);

            if (handledGestures.contains(frame))
                continue;

            framesByTouchCount.insert(geis_frame_touchid_count(frame), frame);
            handledGestures.append(frame);
        }
    }

    for (QMultiMap<unsigned int, GeisFrame>::ConstIterator it = framesByTouchCount.constBegin(); it != framesByTouchCount.constEnd(); it++)
        frames.prepend(*it);
}

QPointF GeisSingleton::pointForTouch(GeisTouch touch)
{
    GeisFloat x, y;

    x = getAttribute<GeisFloat, GeisTouch>(touch, GEIS_TOUCH_ATTRIBUTE_X);
    y = getAttribute<GeisFloat, GeisTouch>(touch, GEIS_TOUCH_ATTRIBUTE_Y);

    return QPointF(x, y);
}

QRectF GeisSingleton::rectForTouches(const QSet<GeisTouch> &touches)
{
    QPointF min;
    QPointF max;
    bool first = true;

    foreach (GeisTouch touch, touches) {
        QPointF point(pointForTouch(touch));

        if (first) {
            min.setX(point.x());
            max.setX(point.x());
            min.setY(point.y());
            max.setY(point.y());
            first = false;
        } else {
            if (point.x() < min.x())
                min.setX(point.x());
            else if (point.x() > max.x())
                max.setX(point.x());
            if (point.y() < min.y())
                min.setY(point.y());
            else if (point.y() > max.y())
                max.setY(point.y());
        }
    }

    return QRectF(min, max);
}

QRectF GeisSingleton::rectForFocus(GeisFrame frame)
{
    GeisFloat x;
    GeisFloat y;

    x = getAttribute<GeisFloat, GeisFrame>(frame, GEIS_GESTURE_ATTRIBUTE_FOCUS_X);
    y = getAttribute<GeisFloat, GeisFrame>(frame, GEIS_GESTURE_ATTRIBUTE_FOCUS_Y);

    return QRectF(x, y, 0, 0);
}

bool GeisSingleton::containsClaimedTouches(const QSet<GeisTouch> &touches, const QSet<GeisTouch> &claimedTouches)
{
    foreach (GeisTouch touch, touches)
        if (claimedTouches.contains(touch))
            return true;

    return false;
}

bool GeisSingleton::containsOtherUnclaimedTouches(const GestureArea *area, const QSet<GeisTouch> &touches, const QSet<GeisTouch> &unclaimedTouches)
{
    foreach (GeisTouch touch, unclaimedTouches) {
        if (touches.contains(touch))
            continue;

        if (area->contains(pointForTouch(touch)))
            return true;
    }

    return false;
}

QGraphicsScene *GeisSingleton::findSceneForTouches(WId window, const QRectF &touches)
{
    foreach (const GestureArea *area, _windowMap.values(window)) {
        QGraphicsScene *scene = area->scene();
        QGraphicsView *view = scene->views().first();
        QRectF bounds(view->mapToGlobal(QPoint(0, 0)), view->size());

        /* touches may be null if it consists of only one point */
        if (touches.isValid()) {
            if (bounds.contains(touches))
                return scene;
        } else if (bounds.contains(touches.topLeft()))
            return scene;
    }

    return NULL;
}

void GeisSingleton::handleGestureBeginEvent(GeisEvent event)
{
    GeisTouchSet touchSet;
    GeisGroupSet groupSet;
    QList<GeisFrame> frames;
    QSet<GeisTouch> claimedTouches;
    QSet<GeisTouch> unclaimedTouches;

    touchSet = getAttribute<GeisTouchSet, GeisEvent>(event, GEIS_EVENT_ATTRIBUTE_TOUCHSET);
    groupSet = getAttribute<GeisGroupSet, GeisEvent>(event, GEIS_EVENT_ATTRIBUTE_GROUPSET);

    getGestureFrameList(groupSet, frames);

    for (GeisSize i = 0; i < geis_touchset_touch_count(touchSet); i++)
        unclaimedTouches.insert(geis_touchset_touch(touchSet, i));

    foreach (GeisFrame frame, frames) {
        GeisInteger deviceId;
        WId window;
        QGraphicsScene *scene;
        bool gestureHandled = false;

        deviceId = getAttribute<GeisInteger, GeisFrame>(frame, GEIS_GESTURE_ATTRIBUTE_DEVICE_ID);
        window = getAttribute<GeisInteger, GeisFrame>(frame, GEIS_GESTURE_ATTRIBUTE_EVENT_WINDOW_ID);

        QSet<GeisTouch> touches;
        for (GeisSize i = 0; i < geis_frame_touchid_count(frame); i++) {
            GeisTouchId id = geis_frame_touchid(frame, i);
            GeisTouch touch = geis_touchset_touch_by_id(touchSet, id);
            touches.insert(touch);
        }

        if (containsClaimedTouches(touches, claimedTouches)) {
            geis_gesture_reject(geis_frame_id(frame));
            continue;
        }

        QList<GestureArea::Primitive> primitives;
        for (QHash<GestureArea::Primitive, GeisGestureClass>::ConstIterator it = _classes.constBegin(); it != _classes.constEnd(); it++)
            if (geis_frame_is_class(frame, it.value()))
                primitives.append(it.key());

        QRectF box;

        if (_deviceTypeMap[deviceId] == Gesture::TouchScreen)
            box = rectForTouches(touches);
        else {
            box = rectForFocus(frame);
        }

        scene = findSceneForTouches(window, box);
        if (!scene) {
            geis_gesture_reject(geis_frame_id(frame));
            continue;
        }

        box.translate(scene->views().first()->mapFromGlobal(QPoint(0, 0)));

        QList<QGraphicsItem *> items;
        if (box.isValid())
            items = scene->items(box);
        else
            items = scene->items(box.topLeft());

        foreach (QGraphicsItem *item, items) {
            GestureArea *area = qobject_cast<GestureArea *>(item);
            if (!area)
                continue;

            if (area->getSubscription()->getDevices() != Gesture::All && area->getSubscription()->getDevices() != _deviceTypeMap[deviceId])
                continue;

            if (!primitives.contains(area->getType()))
                continue;

            if (containsOtherUnclaimedTouches(area, touches, unclaimedTouches))
                continue;

            GestureEvent event(area, frame, touchSet);
            if (area->isGestureEventHandled(event)) {
                GestureArea::Primitive primitive = area->getType();

                primitives.removeAll(primitive);
                gestureHandled = true;

                /* Taps are one-shot, don't keep them around as active. */
                if (primitive != GestureArea::Tap)
                    _activeGestures[geis_frame_id(frame)][primitive] = area;
            }
        }

        if (gestureHandled) {
            /* FIXME: Need a way to accept a gesture frame */
            claimedTouches.unite(touches);
        } else
            geis_gesture_reject(geis_frame_id(frame));
    }
}

void GeisSingleton::handleGestureUpdateEvent(GeisEvent event, bool end)
{
    GeisTouchSet touchSet;
    GeisGroupSet groupSet;
    QList<GeisFrame> frames;

    touchSet = getAttribute<GeisTouchSet, GeisEvent>(event, GEIS_EVENT_ATTRIBUTE_TOUCHSET);
    groupSet = getAttribute<GeisGroupSet, GeisEvent>(event, GEIS_EVENT_ATTRIBUTE_GROUPSET);

    getGestureFrameList(groupSet, frames);

    foreach (GeisFrame frame, frames) {
        GeisGestureId id = geis_frame_id(frame);
        QList<GestureArea::Primitive> primitives;

        for (QHash<GestureArea::Primitive, GeisGestureClass>::ConstIterator it = _classes.constBegin(); it != _classes.constEnd(); it++)
            if (geis_frame_is_class(frame, it.value()))
                primitives.append(it.key());

        /* Tap events only send update events. Handle them like begin events. */
        if (primitives.length() == 1 && primitives.first() == GestureArea::Tap && frames.length() == 1) {
            handleGestureBeginEvent(event);
            return;
        }

        foreach (GestureArea::Primitive primitive, primitives) {
            if (!_activeGestures.contains(id) || !_activeGestures[id].contains(primitive))
                continue;

            GestureArea *area = _activeGestures[geis_frame_id(frame)][primitive];
            GestureEvent event(area, frame, touchSet);
            area->handleGestureUpdateEvent(event, end);

            if (end) {
                _activeGestures[id].remove(primitive);
                if (_activeGestures[id].empty())
                    _activeGestures.remove(id);
            }
        }
    }
}

GeisSingleton::~GeisSingleton()
{
    delete _notifier;

    foreach (GeisGestureClass gestureClass, _classes)
        geis_gesture_class_unref(gestureClass);

    if (_geis)
        geis_delete(_geis);
}
