#ifndef GESTURE_H
#define GESTURE_H

#include "touch.h"

#include <geis/geis.h>

#include <QtDeclarative/QDeclarativeItem>

class GestureArea;

/**
 * A property type used for configuring gesture begin and end conditions based
 * on the number of active touches.
 */

class TouchLimits : public QObject
{
    Q_OBJECT
    Q_DISABLE_COPY(TouchLimits)

    /**
     * This property holds the number of touches required to begin a gesture.
     *
     * The default value is two touches.
     */
    Q_PROPERTY(int start READ getStart WRITE setStart NOTIFY startChanged)

public:
    TouchLimits(QObject *parent=0);

    int getStart() const { return _start; };
    void setStart(int newValue);

signals:
    void startChanged();

private:
    int _start;
};

/**
 * An item describing a gesture subscription.
 */

class Gesture : public QDeclarativeItem
{
    Q_OBJECT
    Q_DISABLE_COPY(Gesture)
    Q_ENUMS(DeviceType)

    /**
     * This property holds the gesture begin and end conditions based on the
     * number of active touches.
     */
    Q_PROPERTY(TouchLimits *touches READ getTouches)

    /**
     * This property holds the device type.
     *
     * Only devices matching the device type will generate events. The default
     * value is All.
     */
    Q_PROPERTY(DeviceType devices READ getDevices WRITE setDevices NOTIFY devicesChanged)

public:
    explicit Gesture(QDeclarativeItem *parent = 0);
    ~Gesture();

    /**
     * An enumeration of classes of multitouch input devices.
     */
    enum DeviceType {
        /**
         * All touchscreens and similar direct input devices.
         *
         * Individual touches are mapped to the screen. Touch locations are
         * provided in screen coordinates.
         */
        TouchScreen,

        /**
         * All touchpads and similar indirect input devices.
         *
         * Individual touches are directed to the location of the associated
         * cursor. Single touch drags that are not accepted as gesture events
         * move the cursor. Touch locations are provided in device coordinates.
         */
        TouchPad,

        /**
         * All independent input devices.
         *
         * Individual touches are directed to the location of the associated
         * cursor. Cursor movement is completely independent of touch input.
         * Touch locations are provided in device coordinates.
         */
        Independent,

        /**
         * All input devices.
         *
         * Touch locations are provided according to the class of device that
         * generated the event.
         */
        All
    };

    TouchLimits *getTouches() { return &_touches; };
    DeviceType getDevices() const { return _devices; };
    void setDevices(DeviceType newValue);

    GeisSubscription subscription() const { return _subscription; };
    void setSubscription(GeisSubscription subscription);
    bool enableSubscription();
    bool disableSubscription();
    void deleteSubscription();

signals:
    void typeChanged();
    void devicesChanged();

private:
    TouchLimits _touches;
    DeviceType _devices;
    GeisSubscription _subscription;
};

QML_DECLARE_TYPE(Gesture)
QML_DECLARE_TYPE(TouchLimits)

#endif // GESTURE_H
