#include "gesturearea.h"
#include "gesture.h"
#include "geissingleton.h"

#include <QGraphicsScene>
#include <QGraphicsView>
#include <QGraphicsWidget>

GestureArea::GestureArea(QDeclarativeItem *parent, Primitive type):
    QDeclarativeItem(parent),
    _type(type),
    _subscription(parent),
    _enabled(true),
    _active(false),
    _windowId(0)
{
    connect(GeisSingleton::instance(), SIGNAL(initialized()), SLOT(geisInitialized()));
    connect(&_subscription, SIGNAL(devicesChanged()), SLOT(subscriptionChanged()));
    connect(_subscription.getTouches(), SIGNAL(startChanged()), SLOT(subscriptionChanged()));
}

QVariant GestureArea::itemChange(GraphicsItemChange change, const QVariant &value)
{
    if (change == QGraphicsItem::ItemSceneHasChanged) {
        _windowId = scene()->views().first()->window()->effectiveWinId();

        if (_windowId) {
            if (GeisSingleton::instance()->isInitialized())
                buildSubscription(_subscription);
        } else
            qCritical("Failed to determine window ID of GestureArea");

    }
    return QDeclarativeItem::itemChange(change, value);
}

void GestureArea::geisInitialized()
{
    if (_windowId)
        buildSubscription(_subscription);
}

#define ADD_FILTER_TERM(filter, facility, term...) \
    if (geis_filter_add_term((filter), (facility), ## term) != GEIS_STATUS_SUCCESS) { \
        qCritical("Failed to add term to filter"); \
        throw; \
    }

void GestureArea::subscriptionChanged() {
    _subscription.deleteSubscription();
    if (_windowId)
        buildSubscription(_subscription);
}

void GestureArea::buildSubscription(Gesture &gesture)
{
    Geis geis = GeisSingleton::instance()->geis();
    GeisStatus status;
    GeisFilter filter;
    GeisSubscription subscription;
    const char *primitive;

    /* Return if already built */
    if (gesture.subscription())
        return;

    filter = geis_filter_new(geis, "GestureArea subscription filter");
    if (!filter) {
        qCritical("Failed to create GeisFilter for subscription");
        return;
    }

    ADD_FILTER_TERM(filter, GEIS_FILTER_REGION, GEIS_REGION_ATTRIBUTE_WINDOWID, GEIS_FILTER_OP_EQ, _windowId, NULL);

    switch(gesture.getDevices()) {
    case Gesture::TouchScreen:
        ADD_FILTER_TERM(filter, GEIS_FILTER_DEVICE, GEIS_DEVICE_ATTRIBUTE_DIRECT_TOUCH, GEIS_FILTER_OP_EQ, GEIS_TRUE, NULL);
        ADD_FILTER_TERM(filter, GEIS_FILTER_DEVICE, GEIS_DEVICE_ATTRIBUTE_INDEPENDENT_TOUCH, GEIS_FILTER_OP_EQ, GEIS_FALSE, NULL);
        break;
    case Gesture::TouchPad:
        ADD_FILTER_TERM(filter, GEIS_FILTER_DEVICE, GEIS_DEVICE_ATTRIBUTE_DIRECT_TOUCH, GEIS_FILTER_OP_EQ, GEIS_FALSE, NULL);
        ADD_FILTER_TERM(filter, GEIS_FILTER_DEVICE, GEIS_DEVICE_ATTRIBUTE_INDEPENDENT_TOUCH, GEIS_FILTER_OP_EQ, GEIS_FALSE, NULL);
        break;
    case Gesture::Independent:
        ADD_FILTER_TERM(filter, GEIS_FILTER_DEVICE, GEIS_DEVICE_ATTRIBUTE_DIRECT_TOUCH, GEIS_FILTER_OP_EQ, GEIS_FALSE, NULL);
        ADD_FILTER_TERM(filter, GEIS_FILTER_DEVICE, GEIS_DEVICE_ATTRIBUTE_INDEPENDENT_TOUCH, GEIS_FILTER_OP_EQ, GEIS_TRUE, NULL);
        break;
    default: /* Gesture::All */
        break;
    }

    switch (_type) {
    case GestureArea::Drag:
        primitive = GEIS_GESTURE_DRAG;
        break;
    case GestureArea::Pinch:
        primitive = GEIS_GESTURE_PINCH;
        break;
    case GestureArea::Rotate:
        primitive = GEIS_GESTURE_ROTATE;
        break;
    case GestureArea::Tap:
        primitive = GEIS_GESTURE_TAP;
        break;
    }

    ADD_FILTER_TERM(filter, GEIS_FILTER_CLASS, GEIS_CLASS_ATTRIBUTE_NAME, GEIS_FILTER_OP_EQ, primitive, NULL);
    ADD_FILTER_TERM(filter, GEIS_FILTER_CLASS, GEIS_GESTURE_ATTRIBUTE_TOUCHES, GEIS_FILTER_OP_EQ, gesture.getTouches()->getStart());

    subscription = geis_subscription_new(geis, "GestureArea subscription", GEIS_SUBSCRIPTION_NONE);
    if (!subscription) {
        qCritical("Failed to create GeisSubscription");
        return;
    }

    status = geis_subscription_add_filter(subscription, filter);
    if (status != GEIS_STATUS_SUCCESS) {
        qCritical("Failed to add filter to subscription");
        geis_subscription_delete(subscription);
        goto error;
    }

    _subscription.setSubscription(subscription);

    if (!_subscription.enableSubscription())
        goto error;

    GeisSingleton::instance()->registerGestureArea(this);

error:
    geis_filter_delete(filter);
}

void GestureArea::setEnabled(bool newValue)
{
    if (_enabled != newValue) {
        if (newValue == true) {
            if (!_subscription.enableSubscription())
                return;
        } else {
            if (!_subscription.disableSubscription())
                return;
        }

        _enabled = newValue;
        emit enabledChanged();
    }
}

bool GestureArea::superIsGestureEventHandled(GestureEvent &event)
{
    float x;
    float y;

    if (_type == Drag || _type == Tap) {
        x = event.getAttributes().value(GEIS_GESTURE_ATTRIBUTE_POSITION_X).toFloat();
        y = event.getAttributes().value(GEIS_GESTURE_ATTRIBUTE_POSITION_Y).toFloat();
    } else {
        x = event.getAttributes().value(GEIS_GESTURE_ATTRIBUTE_CENTROID_X).toFloat();
        y = event.getAttributes().value(GEIS_GESTURE_ATTRIBUTE_CENTROID_Y).toFloat();
    }

    _centroid.set(QPointF(x, y));

    /* Handle instantaneous gestures. */
    if (_type == Tap) {
        emit gestureOccurred(&event);
        return ((_centroid.receivers(SIGNAL(originalChanged())) > 0 ||
                 _centroid.receivers(SIGNAL(currentChanged())) > 0 ||
                 receivers(SIGNAL(gestureOccurred(GestureEvent *))) > 0) && event.getAccepted());
    }

    _active = true;
    emit activeChanged();
    emit gestureStart(&event);
    return ((receivers(SIGNAL(activeChanged())) > 0 ||
             _centroid.receivers(SIGNAL(originalChanged())) > 0 ||
             _centroid.receivers(SIGNAL(currentChanged())) > 0 ||
             receivers(SIGNAL(gestureStart(GestureEvent *))) > 0) && event.getAccepted());
}

void GestureArea::superHandleGestureUpdateEvent(GestureEvent &event, bool end)
{
    float x;
    float y;

    if (_type == GestureArea::Drag) {
        x = event.getAttributes().value(GEIS_GESTURE_ATTRIBUTE_POSITION_X).toFloat();
        y = event.getAttributes().value(GEIS_GESTURE_ATTRIBUTE_POSITION_Y).toFloat();
    } else {
        x = event.getAttributes().value(GEIS_GESTURE_ATTRIBUTE_CENTROID_X).toFloat();
        y = event.getAttributes().value(GEIS_GESTURE_ATTRIBUTE_CENTROID_Y).toFloat();
    }

    _centroid.update(QPointF(x, y));

    if (end) {
        _active = false;
        emit activeChanged();
        emit gestureEnd(&event);
    } else
        emit gestureUpdate(&event);
}
