#ifndef GESTUREAREA_H
#define GESTUREAREA_H

#include <QtDeclarative/QDeclarativeItem>
#include <geis/geis.h>

#include "gesture.h"
#include "gestureevent.h"

class GestureArea;

/**
 * A property type that conveys the original and current value of a variant.
 */

class GestureProperty : public QObject
{
    Q_OBJECT
    Q_DISABLE_COPY(GestureProperty)

    /**
     * This property holds the original value of the variant.
     */
    Q_PROPERTY(QVariant original READ getOriginal NOTIFY originalChanged)

    /**
     * This property holds the current value of the variant.
     */
    Q_PROPERTY(QVariant current READ getCurrent NOTIFY currentChanged)

public:
    GestureProperty(QObject *parent = 0) : QObject(parent) {}

    QVariant getOriginal() const { return _original; };
    QVariant getCurrent() const { return _current; };

    friend class GestureArea;
    friend class PinchArea;
    friend class RotateArea;

signals:
    void originalChanged();
    void currentChanged();

private:
    QVariant _original;
    QVariant _current;

    void set(QVariant newValue) {
        _original = newValue;
        emit originalChanged();
        _current = newValue;
        emit currentChanged();
    };
    void update(QVariant newValue) {
        if (_current != newValue) {
            _current = newValue;
            emit currentChanged();
        }
    };
};

class GestureArea : public QDeclarativeItem
{
    Q_OBJECT
    Q_DISABLE_COPY(GestureArea)

    /**
     * This property holds the subscription for the item.
     *
     * When the subscription is modified, the old subscription is deleted and a
     * new subscription is created.
     */
    Q_PROPERTY(Gesture *subscription READ getSubscription CONSTANT)

    /**
     * This property holds whether the item accepts gesture events.
     *
     * By default, this property is true.
     */
    Q_PROPERTY(bool enabled READ getEnabled WRITE setEnabled NOTIFY enabledChanged)

    /**
     * This property denotes whether the gesture is currently active.
     *
     * The value is only valid during gesture signal processing.
     */
    Q_PROPERTY(bool active READ getActive NOTIFY activeChanged)

    /**
     * This property holds the position of the centroid of the touches
     * comprising the gesture.
     */
    Q_PROPERTY(GestureProperty *centroid READ getCentroid CONSTANT)

public:
    enum Primitive {
        Drag,
        Pinch,
        Rotate,
        Tap
    };

    explicit GestureArea() {}; /* To satisfy QML */
    explicit GestureArea(QDeclarativeItem *parent, Primitive type);

    Primitive getType() const { return _type; };
    Gesture *getSubscription() { return &_subscription; };
    bool getEnabled() const { return _enabled; };
    void setEnabled(bool newValue);
    bool getActive() const { return _active; };
    GestureProperty *getCentroid() { return &_centroid; };
    WId getWindowId() const { return _windowId; };
    virtual bool isGestureEventHandled(GestureEvent &event) = 0;
    virtual void handleGestureUpdateEvent(GestureEvent &event, bool end) = 0;

signals:
    void enabledChanged();
    void activeChanged();

    /**
     * This handler is called when a gesture starts.
     *
     * The gesture parameter provides information about the gesture. This
     * handler is called for non-instantaneous gestures.
     *
     * If the accepted property of the GestureEvent parameter is set to false in the handler,
     * the gesture is rejected and no further signals will be emitted for the
     * gesture. The accepted property defaults to true.
     */
    void gestureStart(GestureEvent *gesture);

    /**
     * This handler is called when a gesture is updated.
     *
     * The gesture parameter provides information about the gesture. This
     * handler is called for non-instantaneous gestures.
     *
     * The accepted property of the GestureEvent parameter is ignored in this
     * handler.
     */
    void gestureUpdate(GestureEvent *gesture);

    /**
     * This handler is called when a gesture ends.
     *
     * The gesture parameter provides information about the gesture. This
     * handler is called for non-instantaneous gestures.
     *
     * The accepted property of the GestureEvent parameter is ignored in this
     * handler.
     */
    void gestureEnd(GestureEvent *gesture);

    /**
     * This handler is called when an instantaneous gesture occurs.
     *
     * The gesture parameter provides information about the gesture. This
     * handler is called for instantaneous gestures.
     *
     * If the accepted property of the GestureEvent parameter is set to false in the handler,
     * the gesture is rejected and no further signals will be emitted for the
     * gesture. The accepted property defaults to true.
     */
    void gestureOccurred(GestureEvent *gesture);

protected:
    Primitive _type;
    Gesture _subscription;
    bool _enabled;
    bool _active;
    GestureProperty _centroid;

    QVariant itemChange(GraphicsItemChange change, const QVariant &value);
    bool superIsGestureEventHandled(GestureEvent &event);
    void superHandleGestureUpdateEvent(GestureEvent &event, bool end);

private:
    void buildSubscription(Gesture &gesture);

    WId _windowId;

private slots:
    void subscriptionChanged();

private slots:
    void geisInitialized();
};

QML_DECLARE_TYPE(GestureArea)
QML_DECLARE_TYPE(GestureProperty)

#endif // GESTUREAREA_H
