#ifndef GESTUREEVENT_H
#define GESTUREEVENT_H

#include <QtDeclarative/QDeclarativeItem>
#include <QtDeclarative/QDeclarativePropertyMap>

#include "gesture.h"
#include "touch.h"

/**
 * An item that provides information about a gesture event.
 */

class GestureEvent : public QDeclarativeItem
{
    Q_OBJECT
    Q_DISABLE_COPY(GestureEvent)

    /**
     * Setting accepted to false in the onGestureBegin handler causes the
     * gesture to be rejected.
     *
     * Other GestureArea items may accept the event, or the event may be
     * rejected entirely, and the touch events may be reprocessed by the system.
     * This property defaults to true.
     */
    Q_PROPERTY(bool accepted READ getAccepted WRITE setAccepted NOTIFY acceptedChanged)

    /**
     * This property holds the list of touches comprising the gesture.
     */
    Q_PROPERTY(const QDeclarativeListProperty<Touch> touches READ getTouches)

    /**
     * This property holds a reference to the gesture subscription.
     */
    Q_PROPERTY(const Gesture *gesture READ getGesture CONSTANT)

    /**
     * This property holds the ID of the gesture event.
     */
    Q_PROPERTY(int id READ getId CONSTANT)

    /**
     * This property holds a variant map of attributes of the gesture.
     *
     * All uTouch gesture attributes are available in this property. The keys
     * may be referenced in the geis.h header file.
     */
    Q_PROPERTY(const QVariantMap attributes READ getAttributes CONSTANT)

public:
    explicit GestureEvent() {}; /* To satisfy QML */
    explicit GestureEvent(GestureArea *area, GeisFrame frame, GeisTouchSet touchSet);

    bool getAccepted() const { return _accepted; };
    const Gesture *getGesture() { return _gesture; };
    int getId() const { return _id; };
    const QVariantMap getAttributes() const { return _attributes; };

    void setAccepted(bool newValue);
    QDeclarativeListProperty<Touch> getTouches() { return QDeclarativeListProperty<Touch>(this, _touches); };

signals:
    void acceptedChanged();

private:
    int _id;
    Gesture *_gesture;
    QList<Touch *> _touches;
    QVariantMap _attributes;
    bool _accepted;
};

QML_DECLARE_TYPE(GestureEvent)

#endif // GESTUREEVENT_H
